#include "ReleasePoint.hpp"
#include <iostream>
#include <functional>
#include <boost/numeric/odeint.hpp>
#include <boost/bind.hpp>

using namespace std;

using namespace boost::numeric::odeint;

runge_kutta4< state_type > stepper;

/* Set physical properties of object dropped
 * The chosen object is a ball
 * r: Radius of the object
 * m: Mass of the object
 * C_D: Drag coefficient of the object (0.47 for a smooth sphere)
 */
double const ReleasePoint::r = 0.05;
double const ReleasePoint::m = 0.2;
double const ReleasePoint::C_D = 0.5;

/* Set the density of air
 * The chosen value (1.269) represents the density of air given
 * 1 atm and +5 degrees Celsius
 */
double const ReleasePoint::rho = 1.269;

/* Set useful constants
 * g: Gravity acceleration on the Earth
 * pi: The mathematical constant
 */
double const ReleasePoint::g = 9.81;
double const ReleasePoint::pi = 3.1415;

ReleasePoint::ReleasePoint(state_type wind_input, state_type x_input , double delta_time)
{
	this->wind = wind_input;
	this->x = x_input;
	this->dt = delta_time;
	this->A = pi*r*r;
	this->b = 0.5*A*C_D*rho;
	this->target_x = x_input[3];
	this->target_y = x_input[4];
}

void ReleasePoint::operator()( const state_type &x , state_type &dxdt , double t )
{
	// Differential equations for a three dimensional free fall
	dxdt[0] = -(this->b/this->m)*(x[0]-wind[0])*sqrt(((x[0]-wind[0])*(x[0]-wind[0]))+((x[1]-wind[1])*(x[1]-wind[1]))+((x[2]-wind[2])*(x[2]-wind[2])));
	dxdt[1] = -(this->b/this->m)*(x[1]-wind[1])*sqrt(((x[0]-wind[0])*(x[0]-wind[0]))+((x[1]-wind[1])*(x[1]-wind[1]))+((x[2]-wind[2])*(x[2]-wind[2])));
	dxdt[2] = -(this->b/this->m)*(x[2]-wind[2])*sqrt(((x[0]-wind[0])*(x[0]-wind[0]))+((x[1]-wind[1])*(x[1]-wind[1]))+((x[2]-wind[2])*(x[2]-wind[2])))-this->g;
	dxdt[3] = x[0];
	dxdt[4] = x[1];
	dxdt[5] = x[2];
}

void ReleasePoint::operator()( const state_type &x , const double t )
{
    cout << t << '\t' << x[0] << '\t' << x[1] << '\t' << x[2] << endl;
    cout << t << '\t' << x[3] << '\t' << x[4] << '\t' << x[5] << endl;

}

state_type ReleasePoint::calc_release_point(double altitude)
{
	// Solve the system of differential equations

    //integrate_const(stepper , ReleasePoint(*this), this->x, 0.0, 3.0 , this->dt , ReleasePoint(*this));
    //for ( double ts = 0.0 ; ts < 3 ; ts+= dt)
	double ts = 0.0;
	while(true)
    {
    	stepper.do_step( ReleasePoint(*this) , this->x , ts , this->dt );
    	//cout << '\t' << x[0] << '\t' << x[1] << '\t' << x[2] << endl;
    	//cout << '\t' << x[3] << '\t' << x[4] << '\t' << x[5] << endl;
    	if( this->x[5] < 0 )
    	{
    		break;
    	}
    	ts += dt;
    }

	cout << '\t' << x[3] << '\t' << x[4] << '\t' << x[5] << '\t' << ts << endl;

	// Calculate position of release in (x,y)-plane
    double dx = this->x[3]-this->target_x;
    double dy = this->x[4]-this->target_y;
    this->x[3] = this->target_x-dx;
    this->x[4] = this->target_y-dy;
	this->x[5] = altitude;



	//cout << '\t' << x[3] << '\t' << x[4] << '\t' << x[5] << endl;

	return this->x;

}

state_type ReleasePoint::get_release_point()
{
	return this->x;
}
