#include "GenerateAugDubins.h"
#include <cmath>
#include <DUNE/DUNE.hpp>


/* Set useful constants
 * g: Gravity acceleration on the Earth
 * pi: The mathematical constant
 * max_bank_angle_deg: Max bank of the UAV in degrees
 */
double const GenerateAugDubins::g = 9.81;
double const GenerateAugDubins::pi = 3.1415;
double const GenerateAugDubins::max_bank_angle_deg = 30.0;

GenerateAugDubins::GenerateAugDubins(bool clockwise, double IAS, double wind, position target, state_type release_point, double displacedCircleDist, position ref_ned_frame)
{
	this->ref = ref_ned_frame;

	deg2rad( );
	calcRotMatrix( clockwise);
	calculateRadius(IAS, wind);
	findCircleCenters(target, release_point, displacedCircleDist);
	findHalfplanes();
	findHalfplanesDirectRoute();
}

void GenerateAugDubins::deg2rad( )
{
	this->max_bank_angle_rad = (this->max_bank_angle_deg/180)*this->pi;
}

void GenerateAugDubins::calcRotMatrix( bool clockwise)
{

	if( clockwise )
	{
		// -90
		this->R_z[0][0] = 0;
		this->R_z[0][1] = 1;
		this->R_z[1][0] = -1;
		this->R_z[1][1] = 0;
	}
	else
	{
		// +90
		this->R_z[0][0] = 0;
		this->R_z[0][1] = -1;
		this->R_z[1][0] = 1;
		this->R_z[1][1] = 0;
	}
}

void GenerateAugDubins::calculateRadius(double IAS, double wind_magnitude)
{
	double max_true_air_speed = IAS + wind_magnitude;

	// Increase max radius by multiplying with the safety_gain
	double safety_gain = 1.1;

	// Calculate max radius based on wind and indicated air speed and increase by multiplying with the safety_gain
	this->R = ((max_true_air_speed*max_true_air_speed)/(this->g*tan(this->max_bank_angle_rad)))*safety_gain;
}

void GenerateAugDubins::findCircleCenters(position target, state_type release_point, double length)
{
	// Circle center c1 is the same as the position of the target
	this->c1 = target;
	this->c2[0] = release_point[3] + this->R*(this->R_z[0][0]*release_point[0] + this->R_z[0][1]*release_point[1]);
	this->c2[1] = release_point[4] + this->R*(this->R_z[1][0]*release_point[0] + this->R_z[1][1]*release_point[1]);
	this->c3[0] = this->c2[0] + length*release_point[0];
	this->c3[1] = this->c2[1] + length*release_point[1];

}

// Calculate the Dubins Path between two circles
void GenerateAugDubins::calculateDubinsPath(position start_circle, position end_circle, halfplane &H1, halfplane &H2)
{
	position c_s = start_circle;
	position c_e = end_circle;

	/*
	 * Calculate H1
	 *
	 *      | z_11   z_12 |
	 * H1 = |             |
	 *      | q_11   q_12 |
	 *      | c_s_x  c_s_y|
	 */

	// Calculate q_11
	H1[1][0] = (c_e[0]-c_s[0])/sqrt((c_e[0]-c_s[0])*(c_e[0]-c_s[0]) + (c_e[1]-c_s[1])*(c_e[1]-c_s[1]));
	// Calculate q_12
	H1[1][1] = (c_e[1]-c_s[1])/sqrt((c_e[0]-c_s[0])*(c_e[0]-c_s[0]) + (c_e[1]-c_s[1])*(c_e[1]-c_s[1]));

	// Calculate z_11
	H1[0][0] = c_s[0] + this->R*(R_z[0][0]*H1[1][0] + R_z[0][1]*H1[1][1]);
	// Calculate z_12
	H1[0][1] = c_s[1] + this->R*(R_z[1][0]*H1[1][0] + R_z[1][1]*H1[1][1]);

	// Include the circle center in the H matrix
	H1[2][0] = c_s[0];
	H1[2][1] = c_s[1];

	// Convert the coordinates from NED to WGS84
	position q1_wgs84 = NEDtoWGS84(H1[0][0], H1[0][1]);
	H1[0][0] = q1_wgs84[0];
	H1[0][1] = q1_wgs84[1];

	position c1_wgs84 = NEDtoWGS84(H1[2][0], H1[2][1]);
	H1[2][0] = c1_wgs84[0];
	H1[2][1] = c1_wgs84[1];


	/*
	 * Calculate H2
	 *
	 *      | z_21   z_22 |
	 * H2 = |             |
	 *      | q_21   q_22 |
	 *      | c_e_x  c_e_y|
	 */

	// q_2 = q_1
	H2[1][0] = H1[1][0];
	H2[1][1] = H1[1][1];

	// Calculate z_21
	H2[0][0] = c_e[0] + this->R*(this->R_z[0][0]*H1[1][0] + this->R_z[0][1]*H1[1][1]);
	// Calculate z_22
	H2[0][1] = c_e[1] + this->R*(this->R_z[1][0]*H1[1][0] + this->R_z[1][1]*H1[1][1]);

	// Include the circle center in the H matrix
	H2[2][0] = c_e[0];
	H2[2][1] = c_e[1];

	// Convert the coordinates from NED to WGS84
	position q2_wgs84 = NEDtoWGS84(H2[0][0], H2[0][1]);
	H2[0][0] = q2_wgs84[0];
	H2[0][1] = q2_wgs84[1];

	position c2_wgs84 = NEDtoWGS84(H2[2][0], H2[2][1]);
	H2[2][0] = c2_wgs84[0];
	H2[2][1] = c2_wgs84[1];

}

void GenerateAugDubins::findHalfplanes()
{
	// c1 to c2
	calculateDubinsPath(this->c1, this->c2, this->H1, this->H2);

	// c2 to c3
	calculateDubinsPath(this->c2, this->c3, this->H3, this->H4);

	// c3 to c2
	calculateDubinsPath(this->c3, this->c2, this->H5, this->H6);



	this->halfplane_list.push_back(this->H1);
	this->halfplane_list.push_back(this->H2);
	this->halfplane_list.push_back(this->H3);
	this->halfplane_list.push_back(this->H4);
	this->halfplane_list.push_back(this->H5);
	this->halfplane_list.push_back(this->H6);

}

position GenerateAugDubins::NEDtoWGS84(double x, double y)
{
	// Created this function to avoid having to reset the reference WGS84 position

	position wgs84_pos;
	wgs84_pos[0] = this->ref[0];
	wgs84_pos[1] = this->ref[1];

	WGS84::displace( x , y , &wgs84_pos[0] , &wgs84_pos[1] );

	return wgs84_pos;

}

std::list< halfplane > * GenerateAugDubins::getHalfplanes()
{

	std::list< halfplane > *Hlist = &this->halfplane_list;
	return Hlist;
}

double GenerateAugDubins::getRadius()
{
	return this->R;
}

void GenerateAugDubins::findHalfplanesDirectRoute()
{
	// c1 to c3
	calculateDubinsPath(this->c1, this->c3, this->H1_DR, this->H2_DR);

	// c3 to c2
	calculateDubinsPath(this->c3, this->c2, this->H3_DR, this->H4_DR);



	this->halfplane_list2.push_back(this->H1_DR);
	this->halfplane_list2.push_back(this->H2_DR);
	this->halfplane_list2.push_back(this->H3_DR);
	this->halfplane_list2.push_back(this->H4_DR);

}

std::list< halfplane > * GenerateAugDubins::getHalfplanesDR()
{

	std::list< halfplane > *Hlist = &this->halfplane_list2;
	return Hlist;
}
