//***************************************************************************
// Copyright 2007-2014 Universidade do Porto - Faculdade de Engenharia      *
// Laboratório de Sistemas e Tecnologia Subaquática (LSTS)                  *
//***************************************************************************
// This file is part of DUNE: Unified Navigation Environment.               *
//                                                                          *
// Commercial Licence Usage                                                 *
// Licencees holding valid commercial DUNE licences may use this file in    *
// accordance with the commercial licence agreement provided with the       *
// Software or, alternatively, in accordance with the terms contained in a  *
// written agreement between you and Universidade do Porto. For licensing   *
// terms, conditions, and further information contact lsts@fe.up.pt.        *
//                                                                          *
// European Union Public Licence - EUPL v.1.1 Usage                         *
// Alternatively, this file may be used under the terms of the EUPL,        *
// Version 1.1 only (the "Licence"), appearing in the file LICENCE.md       *
// included in the packaging of this file. You may not use this work        *
// except in compliance with the Licence. Unless required by applicable     *
// law or agreed to in writing, software distributed under the Licence is   *
// distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF     *
// ANY KIND, either express or implied. See the Licence for the specific    *
// language governing permissions and limitations at                        *
// https://www.lsts.pt/dune/licence.                                        *
//***************************************************************************
// Author: Simen Fuglaas                                                    *
//***************************************************************************

#include "DUNE/ReleasePayloadConfiguration.hpp"
#include "GenerateAugDubins.h"
#include <cmath>

// DUNE headers.
#include <DUNE/DUNE.hpp>

struct GeoPosition
{
    fp64_t lat;
    fp64_t lon;
    fp64_t height;
};

enum ControlState
{
	m_Initialize, m_Wait, m_TrackingCircle, m_Loiter, m_StraightLine, m_Terminate
	/*
	 * Initialize: 			Initialize the target loiter with default radius and chosen direction of flight (CCW/CW)
	 * Wait:				Wait until the release point has been defined and the halfplanes calculated
	 * TrackingCircle: 		Increase radius if needed and track the new circle
	 * Loiter: 				Loiter with chosen radius and previously chosen direction until halfplane is reached
	 * StraightLine: 		Exit loiter, fly straight, and enter new circle when halfplane is reached
	 * Terminate: 			Enter terminate criteria when final halfplane has been reached
	 */
};

enum CalculationMethod{RELEASE_ON_TARGET = 1, STRAIGHT_LINE, DUBINS_PATH, GROUND_STATION};

namespace PrecisionAirdrop
{
	namespace AugmentedDubinsPath
	{
		using DUNE_NAMESPACES;

		struct Task: public DUNE::Tasks::Task
		{
			bool m_received_target;
			bool m_received_e_state;
			bool m_release_point_defined;
			bool m_new_ref_frame;
			bool m_has_released;
			bool m_ready_for_release;
			bool m_clockwise;
			double m_target_lat;
			double m_target_lon;
			double m_target_drop_height;
			double m_ref_lat;
			double m_ref_lon;
			double m_ref_alt;
			double m_alt;
			double m_vx;
			double m_vy;
			double m_vz;
			double m_u;
			double m_v;
			double m_w;
			double m_wind_x;
			double m_wind_y;
			double m_wind_z;
			double IAS;
			//double GS;
			double m_x_pos;
			double m_y_pos;
			double m_z_pos;
			double wind_magnitude;
			double max_turn_radius;
			double m_tolerance;

			ControlState currstate;

			position m_target_pos;

			state_type w;
			state_type m_release_point;

			std::list< halfplane > halfplane_list;

			GeoPosition m_release_pos;
			IMC::TrajectoryPoint m_release_pos_ned;
			IMC::TrajectoryPoint m_current_pos;

			IMC::ControlLoops m_controlLoop;

			LogTxt m_positionLog;
			LogTxt m_logOnPoR;

			//! Constructor.
			//! @param[in] name task name.
			//! @param[in] ctx context.
			Task(const std::string& name, Tasks::Context& ctx):
			DUNE::Tasks::Task(name, ctx)
			{
				// Tell autopilot to control by waypoints
				m_controlLoop.enable = 1;
				m_controlLoop.mask = IMC::CL_PATH;

				bind<IMC::Target>(this);
				bind<IMC::EstimatedState>(this);
				bind<IMC::EstimatedStreamVelocity>(this);
				bind<IMC::IndicatedSpeed>(this);

				m_positionLog.log_start_latitude_longitude_altitude();
				m_logOnPoR.log_start_deviation_from_ideal();

			}

			//! Update internal state with new parameter values.
			void
			onUpdateParameters(void)
			{
			}

			//! Reserve entity identifiers.
			void
			onEntityReservation(void)
			{
			}

			//! Resolve entity names.
			void
			onEntityResolution(void)
			{
			}

			//! Acquire resources.
			void
			onResourceAcquisition(void)
			{
			}

			//! Initialize resources.
			void
			onResourceInitialization(void)
			{
				m_received_target = false;
				m_received_e_state = false;
				m_release_point_defined = false;
				m_new_ref_frame = true;
				m_has_released = false;
				m_ready_for_release = false;
				m_clockwise = false;
				m_target_lat = 0.0;
				m_target_lon = 0.0;
				m_target_drop_height = 0.0;
				m_ref_lat = 0.0;
				m_ref_lon = 0.0;
				m_ref_alt = 0.0;
				m_alt = 0.0;
				m_vx = 0.0;
				m_vy = 0.0;
				m_vz = 0.0;
				m_u = 0.0;
				m_v = 0.0;
				m_w = 0.0;
				m_wind_x = 0.0;
				m_wind_y = 0.0;
				m_wind_z = 0.0;
				IAS = 0.0;
				//GS = 0.0;
				m_x_pos = 0.0;
				m_y_pos = 0.0;
				m_z_pos = 0.0;
				wind_magnitude = 0.0;
			}

			//! Release resources.
			void
			onResourceRelease(void)
			{
			}

			void
			consume(const IMC::EstimatedState* e_state)
			{
				if( e_state->lat != m_ref_lat || e_state->lon != m_ref_lon)
				{
					m_new_ref_frame = true;
					inf("NEW REF FRAME");
				}

				// Extract origin of NED-frame
				m_ref_lat = e_state->lat;
				m_ref_lon = e_state->lon;
				m_ref_alt = e_state->height;

				//err("Reference Frame (rad): %f, %f, %f", m_ref_lat, m_ref_lon, m_ref_alt);
				double ref_lat_deg = m_ref_lat/3.141592*180;
				double ref_lon_deg = m_ref_lon/3.141592*180;
				//inf("Reference Frame (deg): %f, %f", ref_lat_deg, ref_lon_deg);

				// Extract UAV altitude
				m_alt = e_state->alt;

				// Extract Body Velocity
				m_u = e_state->u;
				m_v = e_state->v;
				m_w = e_state->w;

				// Extract Ground Velocity
				m_vx = e_state->vx;
				m_vy = e_state->vy;
				m_vz = e_state->vz;

				// Extract current NED position
				m_current_pos.x = e_state->x;
				m_current_pos.y = e_state->y;
				m_current_pos.z = e_state->z;

				m_received_e_state = true;

				GeoPosition currentPosition = NEDtoWGS84(e_state->x, e_state->y);
				//m_positionLog.log_latitude_longitude_altitude(currentPosition.lat,
				//		currentPosition.lon, e_state->alt);

			}


			void
			consume(const IMC::EstimatedStreamVelocity* stream_vel)
			{
				//err("Wind vector is: [%f, %f, %f]", stream_vel->x, stream_vel->y, stream_vel->z);

				// Extract wind velocity as given by the NED frame
				// Need to switch x and y directions
				m_wind_x = stream_vel->x;
				m_wind_y = stream_vel->y;
				m_wind_z = stream_vel->z;

				wind_magnitude = sqrt(m_wind_x*m_wind_x + m_wind_y*m_wind_y + m_wind_z*m_wind_z);

				w = {{ m_wind_x, m_wind_y, m_wind_z }};

			}

			void
			consume(const IMC::Target* target)
			{

				int dest = target->getDestinationEntity();

				if(dest == DUBINS_PATH)
				{
					war("RECEIVED DUBINS PATH COMMAND");

					inf("Target coordinates (rad) are: %f, %f, %f", target->lat, target->lon, target->z);
					double lat_deg = target->lat/3.141592*180;
					double lon_deg = target->lon/3.141592*180;
					inf("Target coordinates (deg) are: %f, %f", lat_deg, lon_deg);

					m_tolerance = std::atof(target->label.c_str());


					// Check if message has been received
					m_received_target = true;

					// Extract 2D point of impact and altitude of release
					m_target_lat = target->lat;
					m_target_lon = target->lon;
					m_target_drop_height = target->z;

					// Convert WGS-84 coordinates to NED-frame
					WGS84::displacement( m_ref_lat , m_ref_lon , m_ref_alt , m_target_lat , m_target_lon , m_target_drop_height ,
									&m_x_pos , &m_y_pos , &m_z_pos );

					m_target_pos[0] = m_x_pos;
					m_target_pos[1] = m_y_pos;

					inf("Target Position: [%f, %f]", m_x_pos, m_y_pos);

					IMC::Target confirm;
					confirm.label = "confirm";
					confirm.setDestinationEntity(GROUND_STATION);
					dispatch(confirm);

				}

			}


			void
			consume(const IMC::IndicatedSpeed* ind_speed)
			{
				// Extract Indicated Air Speed
				IAS = ind_speed->value;
				//inf("Received indicated speed: %f", IAS);
			}


			void
			generate_augmented_dubins(state_type release_point, position target)
			{
				/*
				// Convert target from WGS84 to NED
				IMC::TrajectoryPoint target_wgs84 = WGS84toNED(m_target_lat , m_target_lon);
				position target;
				target[0] = target_wgs84.x;
				target[1] = target_wgs84.y;
				*/

				position ref_ned_frame;
				ref_ned_frame[0] = m_ref_lat;
				ref_ned_frame[1] = m_ref_lon;

				// Displacement of farthest circle c3 with respect to the circle c2
				double displacedCircleDist = 5*(IAS+wind_magnitude);

				GenerateAugDubins augmented_dubins = GenerateAugDubins(m_clockwise, IAS, wind_magnitude, target, release_point, displacedCircleDist, ref_ned_frame);
				halfplane_list = *augmented_dubins.getHalfplanesDR();
				//double r = augmented_dubins.getRadius();

				//inf("Radius is: %f", r);

				// Iterate list to check the values
				for(std::list<halfplane>::iterator it=halfplane_list.begin(); it != halfplane_list.end(); ++it )
				{
					halfplane H = *it;
					//inf("Halfplane number: %d", it);
					inf("Halfplane position, z: %f, %f", H[0][0], H[0][1]);
					inf("Halfplane vector, q: %f, %f", H[1][0], H[1][1]);
					inf("Associated circle center, c: %f, %f", H[2][0], H[2][1]);
					inf("========================================================");
				}

				max_turn_radius = augmented_dubins.getRadius();

			}


			void
			control_uav()
			{

				// State machine
				switch (currstate)
				{
					case m_Initialize:
					{
						//m_controlLoop.enable = 1;

						// Send loiter position (target) and default radius. Move to m_TrackingCircle

						// Convert target from WGS84 to NED
						IMC::TrajectoryPoint target_wgs84 = WGS84toNED(m_target_lat , m_target_lon);
						position target_pos;
						target_pos[0] = target_wgs84.x;
						target_pos[1] = target_wgs84.y;

						/*
						 * Calculate the release configuration
						 * get_release_point(): returns a vector [6,1] containing
						 * 						the wind unit vector (w_x,w_y,w_z) and
						 * 						the release position (x,y,z)
						 */

						ReleaseConfiguration r_conf = ReleaseConfiguration(m_alt, w, IAS, target_pos);
						m_release_point = r_conf.get_release_point();

						m_release_pos = NEDtoWGS84(m_release_point[3], m_release_point[4]);

						inf("Release Point: [%f %f]", m_release_pos.lat, m_release_pos.lon);

						// Construct message to send release position to Ground Station for logging
						IMC::Target release_pos;

						release_pos.label = "release";
						release_pos.lat = m_release_pos.lat;
						release_pos.lon = m_release_pos.lon;
						release_pos.cog = atan(m_release_point[1]/m_release_point[0]);
						release_pos.setDestinationEntity(GROUND_STATION);

						dispatch(release_pos);


						/*
						 * Generate the list of halfplanes
						 */
						generate_augmented_dubins(m_release_point, target_pos);

						// Dispatch the waypoint list on the IMC
						IMC::DesiredPath loiterTarget;
						//loiter.setSourceEntity(getEntityId());
						loiterTarget.flags = IMC::DesiredPath::FL_NO_Z; // REMEMBER to set flag for CW/CCW loiter
						loiterTarget.flags |= IMC::DesiredPath::FL_START;
						if (!m_clockwise)
						{
							loiterTarget.flags |= IMC::DesiredPath::FL_CCLOCKW;
						}
						loiterTarget.lradius = max_turn_radius;
						loiterTarget.start_lat = m_target_lat;
						loiterTarget.start_lon = m_target_lon;
						loiterTarget.end_lat = m_target_lat;
						loiterTarget.end_lon = m_target_lon;

						war("AUG DUBINS PATH DISPATCHING WAYPOINT");
						dispatch(m_controlLoop);
						dispatch(loiterTarget);

						currstate = m_TrackingCircle;
						inf("Change state to: %d", currstate);

						break;
					}
					case m_Wait:
					{
						// This state desires to ensure that the UAV is on the correct side of the initial circle. When done, move to m_Loiter

						IMC::TrajectoryPoint z;

						// View the next halfplane in the list
						halfplane H_init = halfplane_list.front();

						z = WGS84toNED(H_init[0][0] , H_init[0][1]);

						double q1 = H_init[1][0];
						double q2 = H_init[1][1];

						double H_dist_x = m_current_pos.x - z.x;
						double H_dist_y = m_current_pos.y - z.y;
						double H_cross = H_dist_x*q1 + H_dist_y*q2;

						if( H_cross < 0 )
						{
							// The UAV is on the correct side of the initial circle
							currstate = m_Loiter;
							inf("Change state to: %d", currstate);
						}

						break;
					}
					case m_TrackingCircle:
					{
						// Track initial circle to correct radius before entering state: m_Wait

						// Convert current position to WGS84
						GeoPosition curr_pos_wgs84 = NEDtoWGS84(m_current_pos.x, m_current_pos.y);

						double theta = atan(m_v/m_u);

						double dist = WGS84::distance(curr_pos_wgs84.lat, curr_pos_wgs84.lon, 0, m_target_lat, m_target_lon, 0);

						double difference = abs(dist - max_turn_radius);

						// Positive ang_vel indicates CCW rotation. Negative ang_vel indicates CW rotation.
						double ang_vel = IAS*sin(theta)/dist;

						if(m_clockwise)
						{
							ang_vel = ang_vel*(-1);
						}

						inf("Dist to circle center 1: %f", difference);
						inf("Angular Velocity: %f", ang_vel);

						// Check if the UAV is on the circle
						if( difference < 3 && ang_vel > 0)
						{
							currstate = m_Wait;
							inf("Change state to: %d", currstate);
						}
						break;
					}
					case m_Loiter:
					{
						// Follow the current circle and check for halfplane crossing. Upon occurrence: remove element from list and enter state: m_StraightLine

						IMC::TrajectoryPoint z1;

						// View the next halfplane in the list
						halfplane H1 = halfplane_list.front();

						z1 = WGS84toNED(H1[0][0] , H1[0][1]);

						double q11 = H1[1][0];
						double q12 = H1[1][1];

						// Move z1 tangential one (IAS+wind_magnitude) distance behind
						z1.x = z1.x - 1.1*(IAS+wind_magnitude)*q11;
						z1.y = z1.y - 1.1*(IAS+wind_magnitude)*q12;

						/*
						 * Crossing criteria:
						 * Check (p-z)'q >= 0
						 *
						 * p: Current position of UAV
						 * z: Position of halfplane
						 * q: Unit vector in direction of crossing of halfplane
						 */

						double H1_dist_x = m_current_pos.x - z1.x;
						double H1_dist_y = m_current_pos.y - z1.y;
						double H1_cross = H1_dist_x*q11 + H1_dist_y*q12;

						double dist = sqrt(H1_dist_x*H1_dist_x + H1_dist_y*H1_dist_y);

						if( H1_cross >= 0 && dist < m_tolerance)
						{
							// Leave the circle and enter a straight line

							// Delete H1 from the list
							halfplane_list.pop_front();

							IMC::TrajectoryPoint z2;

							// View the next halfplane in the list
							halfplane H2 = halfplane_list.front();

							z2 = WGS84toNED(H2[0][0] , H2[0][1]);
							double q21 = H2[1][0];
							double q22 = H2[1][1];

							// Set waypoints in WGS84
							GeoPosition wp1; //NEDtoWGS84(z11, z12);
							wp1.lat = H1[0][0];
							wp1.lon = H1[0][1];
							GeoPosition wp2; // = NEDtoWGS84(z21, z22);
							//z2 = WGS84toNED(H2[0][0], H2[0][1]);
							//z2.x = z2.x - q21*IAS;
							//z2.y = z2.y - q22*IAS;
							wp2.lat = H2[0][0];
							wp2.lon = H2[0][1];
							GeoPosition wp3; // = NEDtoWGS84(z21, z22);
							z2.x = z2.x + q21*IAS;
							z2.y = z2.y + q22*IAS;
							wp3 = NEDtoWGS84(z2.x, z2.y);

							// Dispatch the waypoint list on the IMC
							IMC::DesiredPath straight_path;
							//straight_path.setSourceEntity(getEntityId());
							straight_path.flags = IMC::DesiredPath::FL_START;
							straight_path.flags |= IMC::DesiredPath::FL_NO_Z;
							//straight_path.flags |= IMC::DesiredPath::FL_DIRECT;
							//straight_path.start_lat = wp1.lat;
							//straight_path.start_lon = wp1.lon;
							//straight_path.end_lat = wp2.lat;
							//straight_path.end_lon = wp2.lon;
							straight_path.start_lat = wp1.lat;
							straight_path.start_lon = wp1.lon;
							straight_path.end_lat = wp2.lat;
							straight_path.end_lon = wp2.lon;


							war("AUG DUBINS PATH DISPATCHING WAYPOINT");
							dispatch(straight_path);

							currstate = m_StraightLine;
							inf("Change state to: %d", currstate);

						}

						break;
					}
					case m_StraightLine:
					{
						// Follow the current straight line and check for halfplane crossing. Upon occurrence: remove element from list
						// If remaining elements in list equals 1 => enter state Terminate; else => enter state Loiter

						// Termination if only one halfplane remains
						if( halfplane_list.size() == 1 )
						{
							/*
							// Dispatch the final waypoint on the IMC
							IMC::DesiredPath end_path;
							//straight_path.setSourceEntity(getEntityId());
							//end_path.flags = IMC::DesiredPath::FL_START;
							end_path.flags = IMC::DesiredPath::FL_DIRECT;
							end_path.flags |= IMC::DesiredPath::FL_NO_Z;
							//end_path.start_lat = wp1.lat;
							//end_path.start_lon = wp1.lon;
							end_path.end_lat = m_target_lat;
							end_path.end_lon = m_target_lon;

							war("AUG DUBINS PATH DISPATCHING WAYPOINT");
							dispatch(end_path);
							*/

							// Enter state: Terminate
							currstate = m_Terminate;
							inf("Change state to: %d", currstate);

							break;
						}
						IMC::TrajectoryPoint z1;

						// View the next halfplane in the list
						halfplane H1 = halfplane_list.front();

						z1 = WGS84toNED(H1[0][0] , H1[0][1]);

						double q11 = H1[1][0];
						double q12 = H1[1][1];

						// Move z1 tangential one IAS distance behind
						z1.x = z1.x - 1.1*(IAS+wind_magnitude)*q11;
						z1.y = z1.y - 1.1*(IAS+wind_magnitude)*q12;

						/*
						// View the next halfplane in the list
						halfplane H1 = halfplane_list.front();
						double z11 = H1[0][0];
						double z12 = H1[0][1];

						double q11 = H1[1][0];
						double q12 = H1[1][1];
						*/

						/*
						 * Crossing criteria:
						 * Check (p-z)'q >= 0
						 *
						 * p: Current position of UAV
						 * z: Position of halfplane
						 * q: Unit vector in direction of crossing of halfplane
						 */

						double H1_dist_x = m_current_pos.x - z1.x;
						double H1_dist_y = m_current_pos.y - z1.y;
						double H1_cross = H1_dist_x*q11 + H1_dist_y*q12;

						double dist = sqrt(H1_dist_x*H1_dist_x + H1_dist_y*H1_dist_y);

						if( H1_cross >= 0 && dist < m_tolerance)
						{
							// Leave the straight line and enter a circle

							GeoPosition c;
							GeoPosition z;

							// Move first waypoint a little ahead of halfplane position
							// Change to ensure the moved waypoint stays on the circle??
							z1.x = z1.x + q11*2*IAS;
							z1.y = z1.y + q12*2*IAS;

							//z1.x = z1.x + max_turn_radius*cos(3.1415/4);
							//z1.y = z1.y + max_turn_radius*sin(3.1415/4);

							z = NEDtoWGS84(z1.x, z1.y);

							// Delete H1 from the list
							halfplane_list.pop_front();

							// View the next halfplane in the list
							halfplane H2 = halfplane_list.front();
							c.lat = H2[2][0];
							c.lon = H2[2][1];
							//z.lat = H1[0][0];
							//z.lon = H1[0][1];


							// Convert NED-positions to WGS84
							//GeoPosition wpLoiter = NEDtoWGS84(c_x, c_y);

							// Dispatch the waypoint list on the IMC
							IMC::DesiredPath loiter;
							//loiter.setSourceEntity(getEntityId());
							loiter.flags = IMC::DesiredPath::FL_NO_Z; // REMEMBER to set flag for CW/CCW loiter
							loiter.flags |= IMC::DesiredPath::FL_START;
							if (!m_clockwise)
							{
								loiter.flags |= IMC::DesiredPath::FL_CCLOCKW;
							}
							loiter.lradius = max_turn_radius;
							loiter.start_lat = H1[0][0];//z.lat;
							loiter.start_lon = H1[0][1];//z.lon;
							loiter.end_lat = c.lat;
							loiter.end_lon = c.lon;

							//dispatch(m_controlLoop);
							war("AUG DUBINS PATH DISPATCHING WAYPOINT");
							dispatch(loiter);

							currstate = m_Loiter;
							inf("Change state to: %d", currstate);

						}

						break;
					}
					case m_Terminate:
					{
						// Default DesiredPath behavior creates a point further behind for safe loitering

						m_received_target = false;
						m_ready_for_release = true;

						//m_controlLoop.enable = 0;
						//dispatch(m_controlLoop);

						break;
					}
				}
			}

			GeoPosition
			NEDtoWGS84(double x, double y)
			{
				// Created this function to avoid having to reset the reference WGS84 position

				GeoPosition wgs84_pos;
				wgs84_pos.lat = m_ref_lat;
				wgs84_pos.lon = m_ref_lon;

				WGS84::displace( x , y , &wgs84_pos.lat , &wgs84_pos.lon );

				return wgs84_pos;

			}
			IMC::TrajectoryPoint
			WGS84toNED(double lat , double lon)
			{
				// Created this to avoid the height offset

				double alt = m_ref_alt;

				IMC::TrajectoryPoint point;
				WGS84::displacement(m_ref_lat, m_ref_lon, m_ref_alt, lat, lon, alt, &point.x, &point.y, &point.z);

				return point;
			}

			void
			release_object( )
			{
				if(m_new_ref_frame)
				{
					WGS84::displacement(m_ref_lat, m_ref_lon, m_ref_alt, m_release_pos.lat, m_release_pos.lon,
							m_release_pos.height, &m_release_pos_ned.x, &m_release_pos_ned.y, &m_release_pos_ned.z);
					m_new_ref_frame  = false;
					war("Target has been received or the reference frame has changed");
				}
				//inf("Release Pos: x: %f, y: %f", m_release_pos_ned.x, m_release_pos_ned.y);
				//inf("Current Pos: x: %f, y: %f", m_current_pos.x, m_current_pos.y);

				// Calculate distance from current position to release point
				double dist_x = m_release_pos_ned.x - m_current_pos.x;
				double dist_y = m_release_pos_ned.y - m_current_pos.y;
				double dist = sqrt(dist_x*dist_x + dist_y*dist_y);
				//inf("dist_x: %f, dist_y: %f, dist: %f", dist_x, dist_y, dist);

				// Check if the current position is within the given perimeter
				if( (!m_has_released) && (dist < m_tolerance) )
				{
					//Turn on the drop mechanism
					inf("Turn on drop mechanism");
					war("UAV within the given coordinates");
					IMC::PayloadRelease release;
					release.confirm = 0;
					dispatch(release);
					m_has_released = true;
				}
				if(m_has_released && (dist > m_tolerance))
				{
					//Turn off the drop mechanism
					inf("Turn off drop mechanism");
					IMC::PayloadRelease release;
					release.confirm = 1;
					dispatch(release);
					m_has_released = false;
					m_ready_for_release = false;
				}
			}

			void
			release_object_halfplane(GeoPosition release)
			{
				IMC::TrajectoryPoint release_ned;

				WGS84::displacement(m_ref_lat, m_ref_lon, m_ref_alt, release.lat, release.lon,
							release.height, &release_ned.x, &release_ned.y, &release_ned.z);

				// Calculate distance from current position to release point
				double dist_x = m_current_pos.x - release_ned.x;
				double dist_y = m_current_pos.y - release_ned.y;
				double dist = sqrt(dist_x*dist_x + dist_y*dist_y);

				double check_passing = dist_x*(-m_release_point[0]) + dist_y*(-m_release_point[1]);

				// Check if the current position is within the given perimeter
				if( (!m_has_released) && (check_passing >= 0) )
				{
					//Turn on the drop mechanism
					inf("Turn on drop mechanism");
					war("UAV within the given coordinates");
					IMC::PayloadRelease release;
					release.confirm = 0;
					dispatch(release);


					GeoPosition curr_pos = NEDtoWGS84(m_current_pos.x, m_current_pos.y);

					m_logOnPoR.log_deviation_from_ideal(m_current_pos, release_ned,
													"halfplane", m_release_point[0], m_release_point[1],
													m_release_point[2], m_vx, m_vy, m_vz);

					// Construct message to send actual release position to Ground Station for logging
					IMC::Target actual_release_pos;

					actual_release_pos.label = "actual_release";
					actual_release_pos.lat = curr_pos.lat;
					actual_release_pos.lon = curr_pos.lon;
					actual_release_pos.cog = atan(m_vy/m_vx);
					actual_release_pos.setDestinationEntity(GROUND_STATION);

					dispatch(actual_release_pos);


					m_has_released = true;

					// Reset state machine
					currstate = m_Initialize;


				}
				if(m_has_released)
				{
					//Turn off the drop mechanism
					inf("Turn off drop mechanism");
					IMC::PayloadRelease release;
					release.confirm = 1;
					dispatch(release);
					m_has_released = false;
					m_ready_for_release = false;
				}

			}

			//! Main loop.
			void
			onMain(void)
			{

				while (!stopping())
				{
					if(m_received_e_state)
					{
						if(m_received_target)
							{
								//inf("Entering control UAV mode in Dubins Path Method");
								control_uav( );
							}
						if(m_ready_for_release)
						{
							release_object_halfplane( m_release_pos );
						}
					}

				waitForMessages(1.0);
				}
			}
		};
	}
}

DUNE_TASK

