//***************************************************************************
// Copyright 2007-2014 Universidade do Porto - Faculdade de Engenharia      *
// Laboratório de Sistemas e Tecnologia Subaquática (LSTS)                  *
//***************************************************************************
// This file is part of DUNE: Unified Navigation Environment.               *
//                                                                          *
// Commercial Licence Usage                                                 *
// Licencees holding valid commercial DUNE licences may use this file in    *
// accordance with the commercial licence agreement provided with the       *
// Software or, alternatively, in accordance with the terms contained in a  *
// written agreement between you and Universidade do Porto. For licensing   *
// terms, conditions, and further information contact lsts@fe.up.pt.        *
//                                                                          *
// European Union Public Licence - EUPL v.1.1 Usage                         *
// Alternatively, this file may be used under the terms of the EUPL,        *
// Version 1.1 only (the "Licence"), appearing in the file LICENCE.md       *
// included in the packaging of this file. You may not use this work        *
// except in compliance with the Licence. Unless required by applicable     *
// law or agreed to in writing, software distributed under the Licence is   *
// distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF     *
// ANY KIND, either express or implied. See the Licence for the specific    *
// language governing permissions and limitations at                        *
// https://www.lsts.pt/dune/licence.                                        *
//***************************************************************************
// Author: Simen Fuglaas                                                    *
//***************************************************************************

// DUNE headers.
#include <DUNE/DUNE.hpp>

struct GeoPosition
{
    fp64_t lat;
    fp64_t lon;
    fp64_t height;
};

enum CalculationMethod{RELEASE_ON_TARGET = 1, STRAIGHT_LINE, DUBINS_PATH, GROUND_STATION};

namespace PrecisionAirdrop
{
  namespace ReleaseObject
  {
    using DUNE_NAMESPACES;

    struct Task: public DUNE::Tasks::Task
    {
    	double m_tolerance;
    	bool m_target_received;
    	bool m_e_state_received;
    	bool m_new_ref_frame;
    	bool m_has_released;
    	bool passed_halfplane;
    	GeoPosition m_ref_pos;
    	GeoPosition m_target_pos;
    	IMC::TrajectoryPoint m_current_pos;
    	IMC::TrajectoryPoint m_target_pos_ned;
    	LogTxt *m_distanceLog;

      //! Constructor.
      //! @param[in] name task name.
      //! @param[in] ctx context.
      Task(const std::string& name, Tasks::Context& ctx):
        DUNE::Tasks::Task(name, ctx)
      {
    	  bind<IMC::EstimatedState>(this);
    	  bind<IMC::Target>(this);
    	  bind<IMC::PayloadRelease>(this);
    	  m_distanceLog = new LogTxt;
    	  m_distanceLog->log_start_distance_from_target();
      }

      //! Update internal state with new parameter values.
      void
      onUpdateParameters(void)
      {
      }

      //! Reserve entity identifiers.
      void
      onEntityReservation(void)
      {
      }

      //! Resolve entity names.
      void
      onEntityResolution(void)
      {
      }

      //! Acquire resources.
      void
      onResourceAcquisition(void)
      {
      }

      //! Initialize resources.
      void
      onResourceInitialization(void)
      {
    	  m_tolerance = 30.0;
    	  m_target_received = false;
    	  m_e_state_received = false;
    	  m_new_ref_frame = false;
    	  m_has_released = false;
    	  passed_halfplane = false;
      }

      //! Release resources.
      void
      onResourceRelease(void)
      {
      }

      void
      consume(const IMC::EstimatedState* e_state)
      {
    	  if( e_state->lat != m_ref_pos.lat || e_state->lon != m_ref_pos.lon)
    	  {
    		  m_new_ref_frame = true;
    		  inf("NEW REF FRAME: %f, %f", m_ref_pos.lat, m_ref_pos.lon);
    	  }

    	  m_ref_pos.lat = e_state->lat;
    	  m_ref_pos.lon = e_state->lon;
    	  m_ref_pos.height = e_state->height;
    	  //inf("Ref frame: %f, %f", m_ref_pos.lat, m_ref_pos.lon);



    	  m_current_pos.x = e_state->x;
    	  m_current_pos.y = e_state->y;
    	  m_current_pos.z = e_state->z;

    	  m_e_state_received = true;
      }

      void
      consume(const IMC::Target* target)
      {
    	  int dest = target->getDestinationEntity();

    	  if(dest == RELEASE_ON_TARGET)
		  {
    		  war("RECEIVED RELEASE OBJECT COMMAND");

			  m_target_pos.lat = target->lat;
			  m_target_pos.lon = target->lon;
			  m_target_pos.height = target->z;
			  m_tolerance = std::atof(target->label.c_str());


			  IMC::Target confirm;
			  confirm.label = "confirm";
			  confirm.setDestinationEntity(GROUND_STATION);
			  dispatch(confirm);

			  m_target_received = true;
		  }
      }

      void
      consume(const IMC::PayloadRelease* confirm_release)
      {

      }

		void
		release_object( )
		{
			if(m_target_received || m_new_ref_frame)
			{
				WGS84::displacement(m_ref_pos.lat, m_ref_pos.lon, m_ref_pos.height, m_target_pos.lat, m_target_pos.lon,
				m_target_pos.height, &m_target_pos_ned.x, &m_target_pos_ned.y, &m_target_pos_ned.z);
				m_target_received = false;
				m_new_ref_frame  = false;
				war("Target has been received or the reference frame has changed");
			}
			inf("Target Pos: x: %f, y: %f", m_target_pos_ned.x, m_target_pos_ned.y);
			inf("Current Pos: x: %f, y: %f", m_current_pos.x, m_current_pos.y);
			double dist_x = m_target_pos_ned.x - m_current_pos.x;
			double dist_y = m_target_pos_ned.y - m_current_pos.y;
			double dist = sqrt(dist_x*dist_x + dist_y*dist_y);
			//inf("dist_x: %f, dist_y: %f, dist: %f", dist_x, dist_y, dist);

			if((!m_has_released) && (dist < m_tolerance) )
			{
				//Turn on the drop mechanism
				war("UAV within the given coordinates");
				IMC::PayloadRelease release;
				release.confirm = 0;
				dispatch(release);
				m_has_released = 1;
			}
			if(m_has_released && (dist > m_tolerance))
			{
				//Turn off the drop mechanism
				IMC::PayloadRelease release;
				release.confirm = 1;
				dispatch(release);
				m_has_released = 0;
			}

		}

		void
		release_object_halfplane(GeoPosition release)
		{
			IMC::TrajectoryPoint release_ned;

			WGS84::displacement(m_ref_pos.lat, m_ref_pos.lon, m_ref_pos.height, release.lat, release.lon,
						release.height, &release_ned.x, &release_ned.y, &release_ned.z);

			// Calculate distance from current position to release point
			double dist_x = m_current_pos.x - release_ned.x;
			double dist_y = m_current_pos.y - release_ned.y;
			double dist = sqrt(dist_x*dist_x + dist_y*dist_y);

			double check_passing = dist_x*(-1) + dist_y*(0);

			// Check if the UAV is within the desired perimeter
			if( dist <= m_tolerance)
			{
				// Log all points inside the perimeter
				war("Inside perimeter");
				m_distanceLog->log_distance_from_target(m_current_pos, release_ned, "ball");

			}

			// Check if the UAV has passed the halfplane
			if( check_passing >= 0 && !passed_halfplane && dist < m_tolerance)
			{
				passed_halfplane = true;
				// Log the first point after crossing the halfplane
				war("Passed halfplane");
				m_distanceLog->log_distance_from_target(m_current_pos, release_ned, "halfplane");
			}

			// Reset halfplane-passing
			if(dist > m_tolerance)
			{
				passed_halfplane = false;
			}

		}

      //! Main loop.
      void
      onMain(void)
      {
        while (!stopping())
        {
        	if(m_e_state_received)
        	{
        		if(m_target_received)
        		{
        			//inf("Preparing to release object in Release Object Method");
					//release_object( );
        			release_object_halfplane(m_target_pos);
					//m_e_state_received = false;

        		}
        	}
          waitForMessages(1.0);
        }
      }
    };
  }
}

DUNE_TASK
