//***************************************************************************
// Copyright 2007-2014 Universidade do Porto - Faculdade de Engenharia      *
// LaboratÃ³rio de Sistemas e Tecnologia SubaquÃ¡tica (LSTS)                  *
//***************************************************************************
// This file is part of DUNE: Unified Navigation Environment.               *
//                                                                          *
// Commercial Licence Usage                                                 *
// Licencees holding valid commercial DUNE licences may use this file in    *
// accordance with the commercial licence agreement provided with the       *
// Software or, alternatively, in accordance with the terms contained in a  *
// written agreement between you and Universidade do Porto. For licensing   *
// terms, conditions, and further information contact lsts@fe.up.pt.        *
//                                                                          *
// European Union Public Licence - EUPL v.1.1 Usage                         *
// Alternatively, this file may be used under the terms of the EUPL,        *
// Version 1.1 only (the "Licence"), appearing in the file LICENCE.md       *
// included in the packaging of this file. You may not use this work        *
// except in compliance with the Licence. Unless required by applicable     *
// law or agreed to in writing, software distributed under the Licence is   *
// distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF     *
// ANY KIND, either express or implied. See the Licence for the specific    *
// language governing permissions and limitations at                        *
// https://www.lsts.pt/dune/licence.                                        *
//***************************************************************************
// Author: Simen Fuglaas                                                    *
//***************************************************************************

#include "DUNE/ReleasePayloadConfiguration/ReleasePoint.hpp"
#include "DUNE/ReleasePayloadConfiguration/ReleaseVelocity.hpp"
#include "StraightLine/WaypointGenerator.h"
#include <cmath>

// DUNE headers.
#include <DUNE/DUNE.hpp>

struct GeoPosition
{
    fp64_t lat;
    fp64_t lon;
    fp64_t height;
};

enum CalculationMethod{RELEASE_ON_TARGET = 1, STRAIGHT_LINE, DUBINS_PATH};

namespace PrecisionAirdrop
{
	using DUNE_NAMESPACES;

	struct Task: public DUNE::Tasks::Task
	{
		bool m_received_target;
		bool m_received_e_state;
		bool m_release_point_defined;
		bool m_new_ref_frame;
		bool m_has_released;
		bool m_wp1_passed;
		double m_target_lat;
		double m_target_lon;
		double m_target_drop_height;
		double m_ref_lat;
		double m_ref_lon;
		double m_ref_alt;
		double m_alt;
		double m_vx;
		double m_vy;
		double m_vz;
		double m_wind_x;
		double m_wind_y;
		double m_wind_z;
		double IAS;
		//double GS;
		double m_x_pos;
		double m_y_pos;
		double m_z_pos;
		double m_tolerance;

		GeoPosition m_release_pos;
		GeoPosition m_wp1_pos;
		IMC::TrajectoryPoint m_wp1_pos_ned;
		IMC::TrajectoryPoint m_release_pos_ned;
		IMC::TrajectoryPoint m_current_pos;

		IMC::ControlLoops m_controlLoop;

		//! Constructor.
		//! @param[in] name task name.
		//! @param[in] ctx context.
		Task(const std::string& name, Tasks::Context& ctx):
		DUNE::Tasks::Task(name, ctx)
		{
			// Tell autopilot to control by waypoints
			m_controlLoop.enable = 1;
			m_controlLoop.mask = IMC::CL_PATH;

			bind<IMC::Target>(this);
			bind<IMC::EstimatedState>(this);
			bind<IMC::EstimatedStreamVelocity>(this);
			//bind<IMC::IndicatedSpeed>(this);

		}

		//! Update internal state with new parameter values.
		void
		onUpdateParameters(void)
		{
		}

		//! Reserve entity identifiers.
		void
		onEntityReservation(void)
		{
		}

		//! Resolve entity names.
		void
		onEntityResolution(void)
		{
		}

		//! Acquire resources.
		void
		onResourceAcquisition(void)
		{
		}

		//! Initialize resources.
		void
		onResourceInitialization(void)
		{
			m_received_target = false;
			m_received_e_state = false;
			m_release_point_defined = false;
			m_new_ref_frame = true;
			m_has_released = false;
			m_wp1_passed = false;
			m_target_lat = 0.0;
			m_target_lon = 0.0;
			m_target_drop_height = 0.0;
			m_ref_lat = 0.0;
			m_ref_lon = 0.0;
			m_ref_alt = 0.0;
			m_alt = 0.0;
			m_vx = 0.0;
			m_vy = 0.0;
			m_vz = 0.0;
			m_wind_x = 0.0;
			m_wind_y = 0.0;
			m_wind_z = 0.0;
			IAS = 0.0;
			//GS = 0.0;
			m_x_pos = 0.0;
			m_y_pos = 0.0;
			m_z_pos = 0.0;
			m_tolerance = 30.0;
		}

		//! Release resources.
		void
		onResourceRelease(void)
		{
		}

		void
		consume(const IMC::EstimatedState* e_state)
		{
			if( e_state->lat != m_ref_lat || e_state->lon != m_ref_lon)
			{
				m_new_ref_frame = true;
				inf("NEW REF FRAME");
			}

			// Extract origin of NED-frame
			m_ref_lat = e_state->lat;
			m_ref_lon = e_state->lon;
			m_ref_alt = e_state->height;

			//err("Reference Frame (rad): %f, %f, %f", m_ref_lat, m_ref_lon, m_ref_alt);
			//double ref_lat_deg = m_ref_lat/3.141592*180;
			//double ref_lon_deg = m_ref_lon/3.141592*180;
			//inf("Reference Frame (deg): %f, %f", ref_lat_deg, ref_lon_deg);

			// Extract UAV altitude
			m_alt = e_state->alt;

			// Extract Ground Velocity
			//m_vx = e_state->vx;
			//m_vy = e_state->vy;
			//m_vz = e_state->vz;

			// Extract current NED position
			m_current_pos.x = e_state->x;
			m_current_pos.y = e_state->y;
			m_current_pos.z = e_state->z;

			m_received_e_state = true;

		}



		void
		consume(const IMC::EstimatedStreamVelocity* stream_vel)
		{
			//err("Wind vector is: [%f, %f, %f]", stream_vel->x, stream_vel->y, stream_vel->z);

			// REMEMBER to change from negative to positive
			// Changed to negative because of reverse installation of piccolo on Agdenes
			// Extract wind velocity
			m_wind_x = -stream_vel->x;
			m_wind_y = -stream_vel->y;
			m_wind_z = stream_vel->z;

		}

		void
		consume(const IMC::Target* target)
		{
			int dest = target->getDestinationEntity();

			if(dest == STRAIGHT_LINE)
			{
				//inf("Source (DUNE instance) ID is: %d", target->getSource());
				//%inf("Source entity (Task instance) ID is: %d", target->getSourceEntity());
				inf("Target coordinates (rad) are: %f, %f, %f", target->lat, target->lon, target->z);
				double lat_deg = target->lat/3.141592*180;
				double lon_deg = target->lon/3.141592*180;
				inf("Target coordinates (deg) are: %f, %f", lat_deg, lon_deg);

				//
				m_tolerance = std::atof(target->label.c_str());


				// Check if message has been received
				m_received_target = true;

				// Extract 2D point of impact and altitude of release
				m_target_lat = target->lat;
				m_target_lon = target->lon;
				m_target_drop_height = target->z;

				// Convert WGS-84 coordinates to NED-frame
				WGS84::displacement( m_ref_lat , m_ref_lon , m_ref_alt , m_target_lat , m_target_lon , m_target_drop_height ,
								&m_x_pos , &m_y_pos , &m_z_pos );

				inf("Target Position: [%f, %f]", m_x_pos, m_y_pos);

				IMC::Target confirm;
				confirm.label = "confirm";
				dispatch(confirm);
			}

		}


		void
		consume(const IMC::IndicatedSpeed* ind_speed)
		{
			// Extract Indicated Air Speed
			IAS = ind_speed->value;
			inf("Received indicated speed: %f", IAS);
		}



		state_type
		calc_point_of_release( )
		{
			/************************************************************/
			/**************** Calculate Point of Release ****************/
			/************************************************************/

				double wind_magnitude = sqrt(m_wind_x*m_wind_x + m_wind_y*m_wind_y + m_wind_z*m_wind_z);

				// If the wind is insignificantly small, set the wind values
				if( wind_magnitude < 1)
				{
					m_wind_x = 0.01;
					m_wind_y = 0.0;
					m_wind_z = 0.0;
					inf("Insignificant wind detected (%f), using wind vector: [%f, %f, %f]", wind_magnitude, m_wind_x, m_wind_y, m_wind_z);
				}

				state_type w = {{ m_wind_x , m_wind_y , m_wind_z }};

				// Calculate the true ground speed
				//GS = sqrt(m_vx*m_vx + m_vy*m_vy + m_vz*m_vz);

				//war("TAS %f", GS);

				ReleaseVelocity direction = ReleaseVelocity( w , IAS );
				state_type release_vel = direction.calc_release_velocity( );

				// State vector at point of release
				state_type x = {{ release_vel[0] , release_vel[1] , release_vel[2] , m_x_pos , m_y_pos , m_alt }};

				ReleasePoint point = ReleasePoint( w , x , 0.01 );
				state_type release_point = point.calc_release_point( m_alt );

				state_type wind_unit_vector = direction.get_wind_unit_vector( );
				//double wind_magnitude = direction.get_wind_magnitude( );

				release_point[0] = wind_unit_vector[0];
				release_point[1] = wind_unit_vector[1];
				release_point[2] = wind_unit_vector[2];

				inf("Release Point: [%f, %f, %f]",release_point[3],release_point[4],release_point[5]);

				return release_point;
		}

		void
		generate_waypoint_trajectory( )
		{

			/********************************************************/
			/**************** Generate Waypoint List ****************/
			/********************************************************/

				state_type release_point = calc_point_of_release( );

				// Reference NED-frame origin
				waypoint ref_ned_frame = {{ m_ref_lat , m_ref_lon , m_ref_alt }};

				WaypointGenerator wp = WaypointGenerator( release_point , IAS , ref_ned_frame );

				waypointlist wpl = wp.get_wp_list_wgs84( );
				waypointlist wpl_ned = wp.get_wp_list( );
				m_release_point_defined = true;
				m_release_pos.lat = wpl[1][0];
				m_release_pos.lon = wpl[1][1];


				inf("Waypoint 1: [%f, %f, %f]", wpl_ned[0][0], wpl_ned[0][1], wpl_ned[0][2]);
				inf("Waypoint 2: [%f, %f, %f]", wpl_ned[1][0], wpl_ned[1][1], wpl_ned[1][2]);
				inf("Waypoint 3: [%f, %f, %f]", wpl_ned[2][0], wpl_ned[2][1], wpl_ned[2][2]);

				m_wp1_pos.lat = wpl[0][0];
				m_wp1_pos.lon = wpl[0][1];


				// Dispatch the waypoint list on the IMC
				IMC::DesiredPath path_1;
				path_1.setSourceEntity(getEntityId());
				path_1.flags = IMC::DesiredPath::FL_START;
				path_1.flags |= IMC::DesiredPath::FL_NO_Z;
				path_1.start_lat = wpl[0][0];
				path_1.start_lon = wpl[0][1];
				//path_1.start_z = wpl[0][2];

				path_1.end_lat = wpl[1][0];
				path_1.end_lon = wpl[1][1];
				//path_1.end_z = wpl[1][2];

				inf("Path 1 start: lat: %f, lon: %f", path_1.start_lat, path_1.start_lon);
				inf("Path 1 end: lat: %f, lon: %f", path_1.end_lat, path_1.end_lon);

				dispatch(m_controlLoop);
				dispatch(path_1);

				//IMC::DesiredPath path_2;
				//path_2.setSourceEntity(getEntityId());
				//path_2.flags |= IMC::DesiredPath::FL_NO_Z;
				//path_2.end_lat = wpl[2][0];
				//path_2.end_lon = wpl[2][1];
				//path_2.end_z = wpl[2][2];

				//inf("Path 2 end: lat: %f, lon: %f", path_2.end_lat, path_2.end_lon);

				//dispatch(path_2);

		}

		void
		release_object( )
		{
			if(m_new_ref_frame)
			{
				WGS84::displacement(m_ref_lat, m_ref_lon, m_ref_alt, m_release_pos.lat, m_release_pos.lon,
						m_release_pos.height, &m_release_pos_ned.x, &m_release_pos_ned.y, &m_release_pos_ned.z);
				WGS84::displacement(m_ref_lat, m_ref_lon, m_ref_alt, m_wp1_pos.lat, m_wp1_pos.lon,
						m_wp1_pos.height, &m_wp1_pos_ned.x, &m_wp1_pos_ned.y, &m_wp1_pos_ned.z);
				m_new_ref_frame  = false;
				war("Target has been received or the reference frame has changed");
			}
			//inf("Release Pos: x: %f, y: %f", m_release_pos_ned.x, m_release_pos_ned.y);
			//inf("Current Pos: x: %f, y: %f", m_current_pos.x, m_current_pos.y);

			// Calculate distance from current position to release point
			double dist_x = m_release_pos_ned.x - m_current_pos.x;
			double dist_y = m_release_pos_ned.y - m_current_pos.y;
			double dist = sqrt(dist_x*dist_x + dist_y*dist_y);
			//inf("dist_x: %f, dist_y: %f, dist: %f", dist_x, dist_y, dist);

			// Calculate distance from current position to WP1
			double dist_wp1_x = m_wp1_pos_ned.x - m_current_pos.x;
			double dist_wp1_y = m_wp1_pos_ned.y - m_current_pos.y;
			double dist_uav_wp1 = sqrt(dist_wp1_x*dist_wp1_x + dist_wp1_y*dist_wp1_y);

			// Check if the UAV has been to WP1
			if( dist_uav_wp1 < 50 ){
				m_wp1_passed = true;
				inf("We have passed waypoint 1");
			}

			if( m_wp1_passed ){
				// Check if the current position is within the given perimeter
				if( (!m_has_released) && (dist < m_tolerance) )
				{
					//Turn on the drop mechanism
					inf("Turn on drop mechanism");
					war("UAV within the given coordinates");
					IMC::PayloadRelease release;
					release.confirm = 0;
					dispatch(release);
					m_has_released = true;
				}
				if(m_has_released && (dist > m_tolerance))
				{
					//Turn off the drop mechanism
					inf("Turn off drop mechanism");
					IMC::PayloadRelease release;
					release.confirm = 1;
					dispatch(release);
					m_has_released = false;
				}
			}
		}

		//! Main loop.
		void
		onMain(void)
		{

			while (!stopping())
			{
				if(m_received_e_state)
				{
					if(m_received_target)
						{
							generate_waypoint_trajectory();
							m_received_target = false;
						}
					if(m_release_point_defined)
					{
						release_object( );
					}
				}

			waitForMessages(1.0);
			}
		}
	};
}

DUNE_TASK
